/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/* 
 *
 *   Copyright (C) 2008 Uchicago Argonne LLC
 *   See COPYRIGHT notice in top-level directory.
 */

#ifndef AD_BGLOCKLESS_INCLUDE
#define AD_PVFS2_INCLUDE

int ADIOI_BGLOCKLESS_Feature(ADIO_File fd, int flag);

#endif

