/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */
#include ".\examplenode.h"

CExampleNode::CExampleNode(void)
{
}

CExampleNode::~CExampleNode(void)
{
}
