/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_IALLGATHERV( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_IALLGATHERV = PMPI_IALLGATHERV
#pragma weak mpi_iallgatherv__ = PMPI_IALLGATHERV
#pragma weak mpi_iallgatherv_ = PMPI_IALLGATHERV
#pragma weak mpi_iallgatherv = PMPI_IALLGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_IALLGATHERV = pmpi_iallgatherv__
#pragma weak mpi_iallgatherv__ = pmpi_iallgatherv__
#pragma weak mpi_iallgatherv_ = pmpi_iallgatherv__
#pragma weak mpi_iallgatherv = pmpi_iallgatherv__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_IALLGATHERV = pmpi_iallgatherv_
#pragma weak mpi_iallgatherv__ = pmpi_iallgatherv_
#pragma weak mpi_iallgatherv_ = pmpi_iallgatherv_
#pragma weak mpi_iallgatherv = pmpi_iallgatherv_
#else
#pragma weak MPI_IALLGATHERV = pmpi_iallgatherv
#pragma weak mpi_iallgatherv__ = pmpi_iallgatherv
#pragma weak mpi_iallgatherv_ = pmpi_iallgatherv
#pragma weak mpi_iallgatherv = pmpi_iallgatherv
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_IALLGATHERV( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPI_IALLGATHERV = PMPI_IALLGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_iallgatherv__ = pmpi_iallgatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_iallgatherv = pmpi_iallgatherv
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_iallgatherv_ = pmpi_iallgatherv_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_IALLGATHERV  MPI_IALLGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_iallgatherv__  mpi_iallgatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_iallgatherv  mpi_iallgatherv
#else
#pragma _HP_SECONDARY_DEF pmpi_iallgatherv_  mpi_iallgatherv_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_IALLGATHERV as PMPI_IALLGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_iallgatherv__ as pmpi_iallgatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_iallgatherv as pmpi_iallgatherv
#else
#pragma _CRI duplicate mpi_iallgatherv_ as pmpi_iallgatherv_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_IALLGATHERV( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_iallgatherv__ = MPI_IALLGATHERV
#pragma weak mpi_iallgatherv_ = MPI_IALLGATHERV
#pragma weak mpi_iallgatherv = MPI_IALLGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_IALLGATHERV = mpi_iallgatherv__
#pragma weak mpi_iallgatherv_ = mpi_iallgatherv__
#pragma weak mpi_iallgatherv = mpi_iallgatherv__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_IALLGATHERV = mpi_iallgatherv_
#pragma weak mpi_iallgatherv__ = mpi_iallgatherv_
#pragma weak mpi_iallgatherv = mpi_iallgatherv_
#else
#pragma weak MPI_IALLGATHERV = mpi_iallgatherv
#pragma weak mpi_iallgatherv__ = mpi_iallgatherv
#pragma weak mpi_iallgatherv_ = mpi_iallgatherv
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_IALLGATHERV( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_iallgatherv__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_iallgatherv_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_iallgatherv( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_iallgatherv__ = PMPI_IALLGATHERV
#pragma weak pmpi_iallgatherv_ = PMPI_IALLGATHERV
#pragma weak pmpi_iallgatherv = PMPI_IALLGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_IALLGATHERV = pmpi_iallgatherv__
#pragma weak pmpi_iallgatherv_ = pmpi_iallgatherv__
#pragma weak pmpi_iallgatherv = pmpi_iallgatherv__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_IALLGATHERV = pmpi_iallgatherv_
#pragma weak pmpi_iallgatherv__ = pmpi_iallgatherv_
#pragma weak pmpi_iallgatherv = pmpi_iallgatherv_
#else
#pragma weak PMPI_IALLGATHERV = pmpi_iallgatherv
#pragma weak pmpi_iallgatherv__ = pmpi_iallgatherv
#pragma weak pmpi_iallgatherv_ = pmpi_iallgatherv
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_iallgatherv_ PMPI_IALLGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_iallgatherv_ pmpi_iallgatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_iallgatherv_ pmpi_iallgatherv
#else
#define mpi_iallgatherv_ pmpi_iallgatherv_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Iallgatherv
#define MPI_Iallgatherv PMPI_Iallgatherv 

#else

#ifdef F77_NAME_UPPER
#define mpi_iallgatherv_ MPI_IALLGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_iallgatherv_ mpi_iallgatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_iallgatherv_ mpi_iallgatherv
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_iallgatherv_ ( void*v1, MPI_Fint *v2, MPI_Fint *v3, void*v4, MPI_Fint v5[], MPI_Fint v6[], MPI_Fint *v7, MPI_Fint *v8, MPI_Fint *v9, MPI_Fint *ierr ){

#ifndef HAVE_MPI_F_INIT_WORKS_WITH_C
    if (MPIR_F_NeedInit){ mpirinitf_(); MPIR_F_NeedInit = 0; }
#endif
    if (v1 == MPIR_F_MPI_IN_PLACE) v1 = MPI_IN_PLACE;
    *ierr = MPI_Iallgatherv( v1, (int)*v2, (MPI_Datatype)(*v3), v4, v5, v6, (MPI_Datatype)(*v7), (MPI_Comm)(*v8), (MPI_Request *)(v9) );
}
