/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_BARRIER( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_barrier__( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_barrier( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_barrier_( MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_BARRIER = PMPI_BARRIER
#pragma weak mpi_barrier__ = PMPI_BARRIER
#pragma weak mpi_barrier_ = PMPI_BARRIER
#pragma weak mpi_barrier = PMPI_BARRIER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_BARRIER = pmpi_barrier__
#pragma weak mpi_barrier__ = pmpi_barrier__
#pragma weak mpi_barrier_ = pmpi_barrier__
#pragma weak mpi_barrier = pmpi_barrier__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_BARRIER = pmpi_barrier_
#pragma weak mpi_barrier__ = pmpi_barrier_
#pragma weak mpi_barrier_ = pmpi_barrier_
#pragma weak mpi_barrier = pmpi_barrier_
#else
#pragma weak MPI_BARRIER = pmpi_barrier
#pragma weak mpi_barrier__ = pmpi_barrier
#pragma weak mpi_barrier_ = pmpi_barrier
#pragma weak mpi_barrier = pmpi_barrier
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_BARRIER( MPI_Fint *, MPI_Fint * );

#pragma weak MPI_BARRIER = PMPI_BARRIER
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_barrier__( MPI_Fint *, MPI_Fint * );

#pragma weak mpi_barrier__ = pmpi_barrier__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_barrier( MPI_Fint *, MPI_Fint * );

#pragma weak mpi_barrier = pmpi_barrier
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_barrier_( MPI_Fint *, MPI_Fint * );

#pragma weak mpi_barrier_ = pmpi_barrier_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_BARRIER  MPI_BARRIER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_barrier__  mpi_barrier__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_barrier  mpi_barrier
#else
#pragma _HP_SECONDARY_DEF pmpi_barrier_  mpi_barrier_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_BARRIER as PMPI_BARRIER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_barrier__ as pmpi_barrier__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_barrier as pmpi_barrier
#else
#pragma _CRI duplicate mpi_barrier_ as pmpi_barrier_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_BARRIER( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_barrier__( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_barrier( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_barrier_( MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_barrier__ = MPI_BARRIER
#pragma weak mpi_barrier_ = MPI_BARRIER
#pragma weak mpi_barrier = MPI_BARRIER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_BARRIER = mpi_barrier__
#pragma weak mpi_barrier_ = mpi_barrier__
#pragma weak mpi_barrier = mpi_barrier__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_BARRIER = mpi_barrier_
#pragma weak mpi_barrier__ = mpi_barrier_
#pragma weak mpi_barrier = mpi_barrier_
#else
#pragma weak MPI_BARRIER = mpi_barrier
#pragma weak mpi_barrier__ = mpi_barrier
#pragma weak mpi_barrier_ = mpi_barrier
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_BARRIER( MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_barrier__( MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_barrier_( MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_barrier( MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_barrier__ = PMPI_BARRIER
#pragma weak pmpi_barrier_ = PMPI_BARRIER
#pragma weak pmpi_barrier = PMPI_BARRIER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_BARRIER = pmpi_barrier__
#pragma weak pmpi_barrier_ = pmpi_barrier__
#pragma weak pmpi_barrier = pmpi_barrier__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_BARRIER = pmpi_barrier_
#pragma weak pmpi_barrier__ = pmpi_barrier_
#pragma weak pmpi_barrier = pmpi_barrier_
#else
#pragma weak PMPI_BARRIER = pmpi_barrier
#pragma weak pmpi_barrier__ = pmpi_barrier
#pragma weak pmpi_barrier_ = pmpi_barrier
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_barrier_ PMPI_BARRIER
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_barrier_ pmpi_barrier__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_barrier_ pmpi_barrier
#else
#define mpi_barrier_ pmpi_barrier_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Barrier
#define MPI_Barrier PMPI_Barrier 

#else

#ifdef F77_NAME_UPPER
#define mpi_barrier_ MPI_BARRIER
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_barrier_ mpi_barrier__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_barrier_ mpi_barrier
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_barrier_ ( MPI_Fint *v1, MPI_Fint *ierr ){
    *ierr = MPI_Barrier( (MPI_Comm)(*v1) );
}
