/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_RACCUMULATE( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_raccumulate__( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_raccumulate( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_raccumulate_( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_RACCUMULATE = PMPI_RACCUMULATE
#pragma weak mpi_raccumulate__ = PMPI_RACCUMULATE
#pragma weak mpi_raccumulate_ = PMPI_RACCUMULATE
#pragma weak mpi_raccumulate = PMPI_RACCUMULATE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_RACCUMULATE = pmpi_raccumulate__
#pragma weak mpi_raccumulate__ = pmpi_raccumulate__
#pragma weak mpi_raccumulate_ = pmpi_raccumulate__
#pragma weak mpi_raccumulate = pmpi_raccumulate__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_RACCUMULATE = pmpi_raccumulate_
#pragma weak mpi_raccumulate__ = pmpi_raccumulate_
#pragma weak mpi_raccumulate_ = pmpi_raccumulate_
#pragma weak mpi_raccumulate = pmpi_raccumulate_
#else
#pragma weak MPI_RACCUMULATE = pmpi_raccumulate
#pragma weak mpi_raccumulate__ = pmpi_raccumulate
#pragma weak mpi_raccumulate_ = pmpi_raccumulate
#pragma weak mpi_raccumulate = pmpi_raccumulate
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_RACCUMULATE( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPI_RACCUMULATE = PMPI_RACCUMULATE
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_raccumulate__( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_raccumulate__ = pmpi_raccumulate__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_raccumulate( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_raccumulate = pmpi_raccumulate
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_raccumulate_( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_raccumulate_ = pmpi_raccumulate_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_RACCUMULATE  MPI_RACCUMULATE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_raccumulate__  mpi_raccumulate__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_raccumulate  mpi_raccumulate
#else
#pragma _HP_SECONDARY_DEF pmpi_raccumulate_  mpi_raccumulate_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_RACCUMULATE as PMPI_RACCUMULATE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_raccumulate__ as pmpi_raccumulate__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_raccumulate as pmpi_raccumulate
#else
#pragma _CRI duplicate mpi_raccumulate_ as pmpi_raccumulate_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_RACCUMULATE( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_raccumulate__( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_raccumulate( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_raccumulate_( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_raccumulate__ = MPI_RACCUMULATE
#pragma weak mpi_raccumulate_ = MPI_RACCUMULATE
#pragma weak mpi_raccumulate = MPI_RACCUMULATE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_RACCUMULATE = mpi_raccumulate__
#pragma weak mpi_raccumulate_ = mpi_raccumulate__
#pragma weak mpi_raccumulate = mpi_raccumulate__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_RACCUMULATE = mpi_raccumulate_
#pragma weak mpi_raccumulate__ = mpi_raccumulate_
#pragma weak mpi_raccumulate = mpi_raccumulate_
#else
#pragma weak MPI_RACCUMULATE = mpi_raccumulate
#pragma weak mpi_raccumulate__ = mpi_raccumulate
#pragma weak mpi_raccumulate_ = mpi_raccumulate
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_RACCUMULATE( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_raccumulate__( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_raccumulate_( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_raccumulate( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_raccumulate__ = PMPI_RACCUMULATE
#pragma weak pmpi_raccumulate_ = PMPI_RACCUMULATE
#pragma weak pmpi_raccumulate = PMPI_RACCUMULATE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_RACCUMULATE = pmpi_raccumulate__
#pragma weak pmpi_raccumulate_ = pmpi_raccumulate__
#pragma weak pmpi_raccumulate = pmpi_raccumulate__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_RACCUMULATE = pmpi_raccumulate_
#pragma weak pmpi_raccumulate__ = pmpi_raccumulate_
#pragma weak pmpi_raccumulate = pmpi_raccumulate_
#else
#pragma weak PMPI_RACCUMULATE = pmpi_raccumulate
#pragma weak pmpi_raccumulate__ = pmpi_raccumulate
#pragma weak pmpi_raccumulate_ = pmpi_raccumulate
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_raccumulate_ PMPI_RACCUMULATE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_raccumulate_ pmpi_raccumulate__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_raccumulate_ pmpi_raccumulate
#else
#define mpi_raccumulate_ pmpi_raccumulate_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Raccumulate
#define MPI_Raccumulate PMPI_Raccumulate 

#else

#ifdef F77_NAME_UPPER
#define mpi_raccumulate_ MPI_RACCUMULATE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_raccumulate_ mpi_raccumulate__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_raccumulate_ mpi_raccumulate
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_raccumulate_ ( void*v1, MPI_Fint *v2, MPI_Fint *v3, MPI_Fint *v4, MPI_Fint *v5, MPI_Fint *v6, MPI_Fint *v7, MPI_Fint *v8, MPI_Fint *v9, MPI_Fint *v10, MPI_Fint *ierr ){
    *ierr = MPI_Raccumulate( v1, (int)*v2, (MPI_Datatype)(*v3), (int)*v4, (MPI_Aint)*v5, (int)*v6, (MPI_Datatype)(*v7), (MPI_Op)*v8, (MPI_Win)*v9, (MPI_Request *)(v10) );
}
