/* -*- Mode: c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2013 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#ifndef _MPIDI_RECVQ_STATISTICS_H_
#define _MPIDI_RECVQ_STATISTICS_H_

#include "mpidi_common_statistics.h"

/* Do not protect by enable macro - see mpidi_common_statistics.h */
/* from ch3u_recvq.c */
extern uint64_t MPIDI_CH3I_unexpected_recvq_buffer_size;

#endif  /* _MPIDI_RECVQ_STATISTICS_H_ */
