/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPIX_COMM_REENABLE_ANYSOURCE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource_( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPIX_COMM_REENABLE_ANYSOURCE = PMPIX_COMM_REENABLE_ANYSOURCE
#pragma weak mpix_comm_reenable_anysource__ = PMPIX_COMM_REENABLE_ANYSOURCE
#pragma weak mpix_comm_reenable_anysource_ = PMPIX_COMM_REENABLE_ANYSOURCE
#pragma weak mpix_comm_reenable_anysource = PMPIX_COMM_REENABLE_ANYSOURCE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPIX_COMM_REENABLE_ANYSOURCE = pmpix_comm_reenable_anysource__
#pragma weak mpix_comm_reenable_anysource__ = pmpix_comm_reenable_anysource__
#pragma weak mpix_comm_reenable_anysource_ = pmpix_comm_reenable_anysource__
#pragma weak mpix_comm_reenable_anysource = pmpix_comm_reenable_anysource__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPIX_COMM_REENABLE_ANYSOURCE = pmpix_comm_reenable_anysource_
#pragma weak mpix_comm_reenable_anysource__ = pmpix_comm_reenable_anysource_
#pragma weak mpix_comm_reenable_anysource_ = pmpix_comm_reenable_anysource_
#pragma weak mpix_comm_reenable_anysource = pmpix_comm_reenable_anysource_
#else
#pragma weak MPIX_COMM_REENABLE_ANYSOURCE = pmpix_comm_reenable_anysource
#pragma weak mpix_comm_reenable_anysource__ = pmpix_comm_reenable_anysource
#pragma weak mpix_comm_reenable_anysource_ = pmpix_comm_reenable_anysource
#pragma weak mpix_comm_reenable_anysource = pmpix_comm_reenable_anysource
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPIX_COMM_REENABLE_ANYSOURCE( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPIX_COMM_REENABLE_ANYSOURCE = PMPIX_COMM_REENABLE_ANYSOURCE
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource__( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpix_comm_reenable_anysource__ = pmpix_comm_reenable_anysource__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpix_comm_reenable_anysource = pmpix_comm_reenable_anysource
#else
extern FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource_( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpix_comm_reenable_anysource_ = pmpix_comm_reenable_anysource_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPIX_COMM_REENABLE_ANYSOURCE  MPIX_COMM_REENABLE_ANYSOURCE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpix_comm_reenable_anysource__  mpix_comm_reenable_anysource__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpix_comm_reenable_anysource  mpix_comm_reenable_anysource
#else
#pragma _HP_SECONDARY_DEF pmpix_comm_reenable_anysource_  mpix_comm_reenable_anysource_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPIX_COMM_REENABLE_ANYSOURCE as PMPIX_COMM_REENABLE_ANYSOURCE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpix_comm_reenable_anysource__ as pmpix_comm_reenable_anysource__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpix_comm_reenable_anysource as pmpix_comm_reenable_anysource
#else
#pragma _CRI duplicate mpix_comm_reenable_anysource_ as pmpix_comm_reenable_anysource_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPIX_COMM_REENABLE_ANYSOURCE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource_( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpix_comm_reenable_anysource__ = MPIX_COMM_REENABLE_ANYSOURCE
#pragma weak mpix_comm_reenable_anysource_ = MPIX_COMM_REENABLE_ANYSOURCE
#pragma weak mpix_comm_reenable_anysource = MPIX_COMM_REENABLE_ANYSOURCE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPIX_COMM_REENABLE_ANYSOURCE = mpix_comm_reenable_anysource__
#pragma weak mpix_comm_reenable_anysource_ = mpix_comm_reenable_anysource__
#pragma weak mpix_comm_reenable_anysource = mpix_comm_reenable_anysource__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPIX_COMM_REENABLE_ANYSOURCE = mpix_comm_reenable_anysource_
#pragma weak mpix_comm_reenable_anysource__ = mpix_comm_reenable_anysource_
#pragma weak mpix_comm_reenable_anysource = mpix_comm_reenable_anysource_
#else
#pragma weak MPIX_COMM_REENABLE_ANYSOURCE = mpix_comm_reenable_anysource
#pragma weak mpix_comm_reenable_anysource__ = mpix_comm_reenable_anysource
#pragma weak mpix_comm_reenable_anysource_ = mpix_comm_reenable_anysource
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPIX_COMM_REENABLE_ANYSOURCE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpix_comm_reenable_anysource__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpix_comm_reenable_anysource_( MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpix_comm_reenable_anysource( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpix_comm_reenable_anysource__ = PMPIX_COMM_REENABLE_ANYSOURCE
#pragma weak pmpix_comm_reenable_anysource_ = PMPIX_COMM_REENABLE_ANYSOURCE
#pragma weak pmpix_comm_reenable_anysource = PMPIX_COMM_REENABLE_ANYSOURCE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPIX_COMM_REENABLE_ANYSOURCE = pmpix_comm_reenable_anysource__
#pragma weak pmpix_comm_reenable_anysource_ = pmpix_comm_reenable_anysource__
#pragma weak pmpix_comm_reenable_anysource = pmpix_comm_reenable_anysource__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPIX_COMM_REENABLE_ANYSOURCE = pmpix_comm_reenable_anysource_
#pragma weak pmpix_comm_reenable_anysource__ = pmpix_comm_reenable_anysource_
#pragma weak pmpix_comm_reenable_anysource = pmpix_comm_reenable_anysource_
#else
#pragma weak PMPIX_COMM_REENABLE_ANYSOURCE = pmpix_comm_reenable_anysource
#pragma weak pmpix_comm_reenable_anysource__ = pmpix_comm_reenable_anysource
#pragma weak pmpix_comm_reenable_anysource_ = pmpix_comm_reenable_anysource
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpix_comm_reenable_anysource_ PMPIX_COMM_REENABLE_ANYSOURCE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpix_comm_reenable_anysource_ pmpix_comm_reenable_anysource__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpix_comm_reenable_anysource_ pmpix_comm_reenable_anysource
#else
#define mpix_comm_reenable_anysource_ pmpix_comm_reenable_anysource_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPIX_Comm_reenable_anysource
#define MPIX_Comm_reenable_anysource PMPIX_Comm_reenable_anysource 

#else

#ifdef F77_NAME_UPPER
#define mpix_comm_reenable_anysource_ MPIX_COMM_REENABLE_ANYSOURCE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpix_comm_reenable_anysource_ mpix_comm_reenable_anysource__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpix_comm_reenable_anysource_ mpix_comm_reenable_anysource
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpix_comm_reenable_anysource_ ( MPI_Fint *v1, MPI_Fint *v2, MPI_Fint *ierr ){
    *ierr = MPIX_Comm_reenable_anysource( (MPI_Comm)(*v1), v2 );
}
