/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */
#ifndef SMPD_HPC_JS_EXPORTS_H_INCLUDED
#define SMPD_HPC_JS_EXPORTS_H_INCLUDED

/* The MS HPC job scheduler opaque user handle */
typedef void * smpd_hpc_js_handle_t;

#endif /* SMPD_HPC_JS_EXPORTS_H_INCLUDED */
