/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_MPROBE( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_mprobe__( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_mprobe( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_mprobe_( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_MPROBE = PMPI_MPROBE
#pragma weak mpi_mprobe__ = PMPI_MPROBE
#pragma weak mpi_mprobe_ = PMPI_MPROBE
#pragma weak mpi_mprobe = PMPI_MPROBE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_MPROBE = pmpi_mprobe__
#pragma weak mpi_mprobe__ = pmpi_mprobe__
#pragma weak mpi_mprobe_ = pmpi_mprobe__
#pragma weak mpi_mprobe = pmpi_mprobe__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_MPROBE = pmpi_mprobe_
#pragma weak mpi_mprobe__ = pmpi_mprobe_
#pragma weak mpi_mprobe_ = pmpi_mprobe_
#pragma weak mpi_mprobe = pmpi_mprobe_
#else
#pragma weak MPI_MPROBE = pmpi_mprobe
#pragma weak mpi_mprobe__ = pmpi_mprobe
#pragma weak mpi_mprobe_ = pmpi_mprobe
#pragma weak mpi_mprobe = pmpi_mprobe
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_MPROBE( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPI_MPROBE = PMPI_MPROBE
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_mprobe__( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_mprobe__ = pmpi_mprobe__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_mprobe( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_mprobe = pmpi_mprobe
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_mprobe_( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_mprobe_ = pmpi_mprobe_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_MPROBE  MPI_MPROBE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_mprobe__  mpi_mprobe__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_mprobe  mpi_mprobe
#else
#pragma _HP_SECONDARY_DEF pmpi_mprobe_  mpi_mprobe_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_MPROBE as PMPI_MPROBE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_mprobe__ as pmpi_mprobe__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_mprobe as pmpi_mprobe
#else
#pragma _CRI duplicate mpi_mprobe_ as pmpi_mprobe_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_MPROBE( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_mprobe__( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_mprobe( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_mprobe_( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_mprobe__ = MPI_MPROBE
#pragma weak mpi_mprobe_ = MPI_MPROBE
#pragma weak mpi_mprobe = MPI_MPROBE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_MPROBE = mpi_mprobe__
#pragma weak mpi_mprobe_ = mpi_mprobe__
#pragma weak mpi_mprobe = mpi_mprobe__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_MPROBE = mpi_mprobe_
#pragma weak mpi_mprobe__ = mpi_mprobe_
#pragma weak mpi_mprobe = mpi_mprobe_
#else
#pragma weak MPI_MPROBE = mpi_mprobe
#pragma weak mpi_mprobe__ = mpi_mprobe
#pragma weak mpi_mprobe_ = mpi_mprobe
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_MPROBE( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_mprobe__( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_mprobe_( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_mprobe( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_mprobe__ = PMPI_MPROBE
#pragma weak pmpi_mprobe_ = PMPI_MPROBE
#pragma weak pmpi_mprobe = PMPI_MPROBE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_MPROBE = pmpi_mprobe__
#pragma weak pmpi_mprobe_ = pmpi_mprobe__
#pragma weak pmpi_mprobe = pmpi_mprobe__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_MPROBE = pmpi_mprobe_
#pragma weak pmpi_mprobe__ = pmpi_mprobe_
#pragma weak pmpi_mprobe = pmpi_mprobe_
#else
#pragma weak PMPI_MPROBE = pmpi_mprobe
#pragma weak pmpi_mprobe__ = pmpi_mprobe
#pragma weak pmpi_mprobe_ = pmpi_mprobe
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_mprobe_ PMPI_MPROBE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_mprobe_ pmpi_mprobe__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_mprobe_ pmpi_mprobe
#else
#define mpi_mprobe_ pmpi_mprobe_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Mprobe
#define MPI_Mprobe PMPI_Mprobe 

#else

#ifdef F77_NAME_UPPER
#define mpi_mprobe_ MPI_MPROBE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_mprobe_ mpi_mprobe__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_mprobe_ mpi_mprobe
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_mprobe_ ( MPI_Fint *v1, MPI_Fint *v2, MPI_Fint *v3, MPI_Fint *v4, MPI_Fint *v5, MPI_Fint *ierr ){

#ifndef HAVE_MPI_F_INIT_WORKS_WITH_C
    if (MPIR_F_NeedInit){ mpirinitf_(); MPIR_F_NeedInit = 0; }
#endif

    if (v5 == MPI_F_STATUS_IGNORE) { v5 = (MPI_Fint*)MPI_STATUS_IGNORE; }
    *ierr = MPI_Mprobe( (int)*v1, (int)*v2, (MPI_Comm)(*v3), (MPI_Message *)(v4), (MPI_Status *)v5 );
}
