/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_IGATHERV( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_igatherv__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_igatherv( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_igatherv_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_IGATHERV = PMPI_IGATHERV
#pragma weak mpi_igatherv__ = PMPI_IGATHERV
#pragma weak mpi_igatherv_ = PMPI_IGATHERV
#pragma weak mpi_igatherv = PMPI_IGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_IGATHERV = pmpi_igatherv__
#pragma weak mpi_igatherv__ = pmpi_igatherv__
#pragma weak mpi_igatherv_ = pmpi_igatherv__
#pragma weak mpi_igatherv = pmpi_igatherv__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_IGATHERV = pmpi_igatherv_
#pragma weak mpi_igatherv__ = pmpi_igatherv_
#pragma weak mpi_igatherv_ = pmpi_igatherv_
#pragma weak mpi_igatherv = pmpi_igatherv_
#else
#pragma weak MPI_IGATHERV = pmpi_igatherv
#pragma weak mpi_igatherv__ = pmpi_igatherv
#pragma weak mpi_igatherv_ = pmpi_igatherv
#pragma weak mpi_igatherv = pmpi_igatherv
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_IGATHERV( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPI_IGATHERV = PMPI_IGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_igatherv__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_igatherv__ = pmpi_igatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_igatherv( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_igatherv = pmpi_igatherv
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_igatherv_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_igatherv_ = pmpi_igatherv_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_IGATHERV  MPI_IGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_igatherv__  mpi_igatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_igatherv  mpi_igatherv
#else
#pragma _HP_SECONDARY_DEF pmpi_igatherv_  mpi_igatherv_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_IGATHERV as PMPI_IGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_igatherv__ as pmpi_igatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_igatherv as pmpi_igatherv
#else
#pragma _CRI duplicate mpi_igatherv_ as pmpi_igatherv_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_IGATHERV( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_igatherv__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_igatherv( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_igatherv_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_igatherv__ = MPI_IGATHERV
#pragma weak mpi_igatherv_ = MPI_IGATHERV
#pragma weak mpi_igatherv = MPI_IGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_IGATHERV = mpi_igatherv__
#pragma weak mpi_igatherv_ = mpi_igatherv__
#pragma weak mpi_igatherv = mpi_igatherv__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_IGATHERV = mpi_igatherv_
#pragma weak mpi_igatherv__ = mpi_igatherv_
#pragma weak mpi_igatherv = mpi_igatherv_
#else
#pragma weak MPI_IGATHERV = mpi_igatherv
#pragma weak mpi_igatherv__ = mpi_igatherv
#pragma weak mpi_igatherv_ = mpi_igatherv
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_IGATHERV( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_igatherv__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_igatherv_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_igatherv( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint [], MPI_Fint [], MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_igatherv__ = PMPI_IGATHERV
#pragma weak pmpi_igatherv_ = PMPI_IGATHERV
#pragma weak pmpi_igatherv = PMPI_IGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_IGATHERV = pmpi_igatherv__
#pragma weak pmpi_igatherv_ = pmpi_igatherv__
#pragma weak pmpi_igatherv = pmpi_igatherv__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_IGATHERV = pmpi_igatherv_
#pragma weak pmpi_igatherv__ = pmpi_igatherv_
#pragma weak pmpi_igatherv = pmpi_igatherv_
#else
#pragma weak PMPI_IGATHERV = pmpi_igatherv
#pragma weak pmpi_igatherv__ = pmpi_igatherv
#pragma weak pmpi_igatherv_ = pmpi_igatherv
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_igatherv_ PMPI_IGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_igatherv_ pmpi_igatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_igatherv_ pmpi_igatherv
#else
#define mpi_igatherv_ pmpi_igatherv_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Igatherv
#define MPI_Igatherv PMPI_Igatherv 

#else

#ifdef F77_NAME_UPPER
#define mpi_igatherv_ MPI_IGATHERV
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_igatherv_ mpi_igatherv__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_igatherv_ mpi_igatherv
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_igatherv_ ( void*v1, MPI_Fint *v2, MPI_Fint *v3, void*v4, MPI_Fint v5[], MPI_Fint v6[], MPI_Fint *v7, MPI_Fint *v8, MPI_Fint *v9, MPI_Fint *v10, MPI_Fint *ierr ){

#ifndef HAVE_MPI_F_INIT_WORKS_WITH_C
    if (MPIR_F_NeedInit){ mpirinitf_(); MPIR_F_NeedInit = 0; }
#endif
    if (v1 == MPIR_F_MPI_IN_PLACE) v1 = MPI_IN_PLACE;
    *ierr = MPI_Igatherv( v1, (int)*v2, (MPI_Datatype)(*v3), v4, v5, v6, (MPI_Datatype)(*v7), (int)*v8, (MPI_Comm)(*v9), (MPI_Request *)(v10) );
}
