/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_ISCATTER( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iscatter__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iscatter( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iscatter_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_ISCATTER = PMPI_ISCATTER
#pragma weak mpi_iscatter__ = PMPI_ISCATTER
#pragma weak mpi_iscatter_ = PMPI_ISCATTER
#pragma weak mpi_iscatter = PMPI_ISCATTER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_ISCATTER = pmpi_iscatter__
#pragma weak mpi_iscatter__ = pmpi_iscatter__
#pragma weak mpi_iscatter_ = pmpi_iscatter__
#pragma weak mpi_iscatter = pmpi_iscatter__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_ISCATTER = pmpi_iscatter_
#pragma weak mpi_iscatter__ = pmpi_iscatter_
#pragma weak mpi_iscatter_ = pmpi_iscatter_
#pragma weak mpi_iscatter = pmpi_iscatter_
#else
#pragma weak MPI_ISCATTER = pmpi_iscatter
#pragma weak mpi_iscatter__ = pmpi_iscatter
#pragma weak mpi_iscatter_ = pmpi_iscatter
#pragma weak mpi_iscatter = pmpi_iscatter
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_ISCATTER( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPI_ISCATTER = PMPI_ISCATTER
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_iscatter__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_iscatter__ = pmpi_iscatter__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_iscatter( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_iscatter = pmpi_iscatter
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_iscatter_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_iscatter_ = pmpi_iscatter_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_ISCATTER  MPI_ISCATTER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_iscatter__  mpi_iscatter__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_iscatter  mpi_iscatter
#else
#pragma _HP_SECONDARY_DEF pmpi_iscatter_  mpi_iscatter_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_ISCATTER as PMPI_ISCATTER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_iscatter__ as pmpi_iscatter__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_iscatter as pmpi_iscatter
#else
#pragma _CRI duplicate mpi_iscatter_ as pmpi_iscatter_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_ISCATTER( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iscatter__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iscatter( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_iscatter_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_iscatter__ = MPI_ISCATTER
#pragma weak mpi_iscatter_ = MPI_ISCATTER
#pragma weak mpi_iscatter = MPI_ISCATTER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_ISCATTER = mpi_iscatter__
#pragma weak mpi_iscatter_ = mpi_iscatter__
#pragma weak mpi_iscatter = mpi_iscatter__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_ISCATTER = mpi_iscatter_
#pragma weak mpi_iscatter__ = mpi_iscatter_
#pragma weak mpi_iscatter = mpi_iscatter_
#else
#pragma weak MPI_ISCATTER = mpi_iscatter
#pragma weak mpi_iscatter__ = mpi_iscatter
#pragma weak mpi_iscatter_ = mpi_iscatter
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_ISCATTER( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_iscatter__( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_iscatter_( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_iscatter( void*, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_iscatter__ = PMPI_ISCATTER
#pragma weak pmpi_iscatter_ = PMPI_ISCATTER
#pragma weak pmpi_iscatter = PMPI_ISCATTER
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_ISCATTER = pmpi_iscatter__
#pragma weak pmpi_iscatter_ = pmpi_iscatter__
#pragma weak pmpi_iscatter = pmpi_iscatter__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_ISCATTER = pmpi_iscatter_
#pragma weak pmpi_iscatter__ = pmpi_iscatter_
#pragma weak pmpi_iscatter = pmpi_iscatter_
#else
#pragma weak PMPI_ISCATTER = pmpi_iscatter
#pragma weak pmpi_iscatter__ = pmpi_iscatter
#pragma weak pmpi_iscatter_ = pmpi_iscatter
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_iscatter_ PMPI_ISCATTER
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_iscatter_ pmpi_iscatter__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_iscatter_ pmpi_iscatter
#else
#define mpi_iscatter_ pmpi_iscatter_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Iscatter
#define MPI_Iscatter PMPI_Iscatter 

#else

#ifdef F77_NAME_UPPER
#define mpi_iscatter_ MPI_ISCATTER
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_iscatter_ mpi_iscatter__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_iscatter_ mpi_iscatter
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_iscatter_ ( void*v1, MPI_Fint *v2, MPI_Fint *v3, void*v4, MPI_Fint *v5, MPI_Fint *v6, MPI_Fint *v7, MPI_Fint *v8, MPI_Fint *v9, MPI_Fint *ierr ){

#ifndef HAVE_MPI_F_INIT_WORKS_WITH_C
    if (MPIR_F_NeedInit){ mpirinitf_(); MPIR_F_NeedInit = 0; }
#endif
    if (v4 == MPIR_F_MPI_IN_PLACE) v4 = MPI_IN_PLACE;
    *ierr = MPI_Iscatter( v1, (int)*v2, (MPI_Datatype)(*v3), v4, (int)*v5, (MPI_Datatype)(*v6), (int)*v7, (MPI_Comm)(*v8), (MPI_Request *)(v9) );
}
