/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_COMM_SPLIT_TYPE( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type__( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type_( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_COMM_SPLIT_TYPE = PMPI_COMM_SPLIT_TYPE
#pragma weak mpi_comm_split_type__ = PMPI_COMM_SPLIT_TYPE
#pragma weak mpi_comm_split_type_ = PMPI_COMM_SPLIT_TYPE
#pragma weak mpi_comm_split_type = PMPI_COMM_SPLIT_TYPE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_COMM_SPLIT_TYPE = pmpi_comm_split_type__
#pragma weak mpi_comm_split_type__ = pmpi_comm_split_type__
#pragma weak mpi_comm_split_type_ = pmpi_comm_split_type__
#pragma weak mpi_comm_split_type = pmpi_comm_split_type__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_COMM_SPLIT_TYPE = pmpi_comm_split_type_
#pragma weak mpi_comm_split_type__ = pmpi_comm_split_type_
#pragma weak mpi_comm_split_type_ = pmpi_comm_split_type_
#pragma weak mpi_comm_split_type = pmpi_comm_split_type_
#else
#pragma weak MPI_COMM_SPLIT_TYPE = pmpi_comm_split_type
#pragma weak mpi_comm_split_type__ = pmpi_comm_split_type
#pragma weak mpi_comm_split_type_ = pmpi_comm_split_type
#pragma weak mpi_comm_split_type = pmpi_comm_split_type
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_COMM_SPLIT_TYPE( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPI_COMM_SPLIT_TYPE = PMPI_COMM_SPLIT_TYPE
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type__( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_comm_split_type__ = pmpi_comm_split_type__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_comm_split_type = pmpi_comm_split_type
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type_( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_comm_split_type_ = pmpi_comm_split_type_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_COMM_SPLIT_TYPE  MPI_COMM_SPLIT_TYPE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_comm_split_type__  mpi_comm_split_type__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_comm_split_type  mpi_comm_split_type
#else
#pragma _HP_SECONDARY_DEF pmpi_comm_split_type_  mpi_comm_split_type_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_COMM_SPLIT_TYPE as PMPI_COMM_SPLIT_TYPE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_comm_split_type__ as pmpi_comm_split_type__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_comm_split_type as pmpi_comm_split_type
#else
#pragma _CRI duplicate mpi_comm_split_type_ as pmpi_comm_split_type_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_COMM_SPLIT_TYPE( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type__( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type_( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_comm_split_type__ = MPI_COMM_SPLIT_TYPE
#pragma weak mpi_comm_split_type_ = MPI_COMM_SPLIT_TYPE
#pragma weak mpi_comm_split_type = MPI_COMM_SPLIT_TYPE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_COMM_SPLIT_TYPE = mpi_comm_split_type__
#pragma weak mpi_comm_split_type_ = mpi_comm_split_type__
#pragma weak mpi_comm_split_type = mpi_comm_split_type__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_COMM_SPLIT_TYPE = mpi_comm_split_type_
#pragma weak mpi_comm_split_type__ = mpi_comm_split_type_
#pragma weak mpi_comm_split_type = mpi_comm_split_type_
#else
#pragma weak MPI_COMM_SPLIT_TYPE = mpi_comm_split_type
#pragma weak mpi_comm_split_type__ = mpi_comm_split_type
#pragma weak mpi_comm_split_type_ = mpi_comm_split_type
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_COMM_SPLIT_TYPE( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_comm_split_type__( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_comm_split_type_( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_comm_split_type( MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_comm_split_type__ = PMPI_COMM_SPLIT_TYPE
#pragma weak pmpi_comm_split_type_ = PMPI_COMM_SPLIT_TYPE
#pragma weak pmpi_comm_split_type = PMPI_COMM_SPLIT_TYPE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_COMM_SPLIT_TYPE = pmpi_comm_split_type__
#pragma weak pmpi_comm_split_type_ = pmpi_comm_split_type__
#pragma weak pmpi_comm_split_type = pmpi_comm_split_type__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_COMM_SPLIT_TYPE = pmpi_comm_split_type_
#pragma weak pmpi_comm_split_type__ = pmpi_comm_split_type_
#pragma weak pmpi_comm_split_type = pmpi_comm_split_type_
#else
#pragma weak PMPI_COMM_SPLIT_TYPE = pmpi_comm_split_type
#pragma weak pmpi_comm_split_type__ = pmpi_comm_split_type
#pragma weak pmpi_comm_split_type_ = pmpi_comm_split_type
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_comm_split_type_ PMPI_COMM_SPLIT_TYPE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_comm_split_type_ pmpi_comm_split_type__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_comm_split_type_ pmpi_comm_split_type
#else
#define mpi_comm_split_type_ pmpi_comm_split_type_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Comm_split_type
#define MPI_Comm_split_type PMPI_Comm_split_type 

#else

#ifdef F77_NAME_UPPER
#define mpi_comm_split_type_ MPI_COMM_SPLIT_TYPE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_comm_split_type_ mpi_comm_split_type__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_comm_split_type_ mpi_comm_split_type
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_comm_split_type_ ( MPI_Fint *v1, MPI_Fint *v2, MPI_Fint *v3, MPI_Fint *v4, MPI_Fint *v5, MPI_Fint *ierr ){
    *ierr = MPI_Comm_split_type( (MPI_Comm)(*v1), (int)*v2, (int)*v3, (MPI_Info)(*v4), (MPI_Comm *)(v5) );
}
