/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_FINALIZED( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_finalized__( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_finalized( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_finalized_( MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_FINALIZED = PMPI_FINALIZED
#pragma weak mpi_finalized__ = PMPI_FINALIZED
#pragma weak mpi_finalized_ = PMPI_FINALIZED
#pragma weak mpi_finalized = PMPI_FINALIZED
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_FINALIZED = pmpi_finalized__
#pragma weak mpi_finalized__ = pmpi_finalized__
#pragma weak mpi_finalized_ = pmpi_finalized__
#pragma weak mpi_finalized = pmpi_finalized__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_FINALIZED = pmpi_finalized_
#pragma weak mpi_finalized__ = pmpi_finalized_
#pragma weak mpi_finalized_ = pmpi_finalized_
#pragma weak mpi_finalized = pmpi_finalized_
#else
#pragma weak MPI_FINALIZED = pmpi_finalized
#pragma weak mpi_finalized__ = pmpi_finalized
#pragma weak mpi_finalized_ = pmpi_finalized
#pragma weak mpi_finalized = pmpi_finalized
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_FINALIZED( MPI_Fint *, MPI_Fint * );

#pragma weak MPI_FINALIZED = PMPI_FINALIZED
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_finalized__( MPI_Fint *, MPI_Fint * );

#pragma weak mpi_finalized__ = pmpi_finalized__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_finalized( MPI_Fint *, MPI_Fint * );

#pragma weak mpi_finalized = pmpi_finalized
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_finalized_( MPI_Fint *, MPI_Fint * );

#pragma weak mpi_finalized_ = pmpi_finalized_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_FINALIZED  MPI_FINALIZED
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_finalized__  mpi_finalized__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_finalized  mpi_finalized
#else
#pragma _HP_SECONDARY_DEF pmpi_finalized_  mpi_finalized_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_FINALIZED as PMPI_FINALIZED
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_finalized__ as pmpi_finalized__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_finalized as pmpi_finalized
#else
#pragma _CRI duplicate mpi_finalized_ as pmpi_finalized_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_FINALIZED( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_finalized__( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_finalized( MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_finalized_( MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_finalized__ = MPI_FINALIZED
#pragma weak mpi_finalized_ = MPI_FINALIZED
#pragma weak mpi_finalized = MPI_FINALIZED
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_FINALIZED = mpi_finalized__
#pragma weak mpi_finalized_ = mpi_finalized__
#pragma weak mpi_finalized = mpi_finalized__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_FINALIZED = mpi_finalized_
#pragma weak mpi_finalized__ = mpi_finalized_
#pragma weak mpi_finalized = mpi_finalized_
#else
#pragma weak MPI_FINALIZED = mpi_finalized
#pragma weak mpi_finalized__ = mpi_finalized
#pragma weak mpi_finalized_ = mpi_finalized
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_FINALIZED( MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_finalized__( MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_finalized_( MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_finalized( MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_finalized__ = PMPI_FINALIZED
#pragma weak pmpi_finalized_ = PMPI_FINALIZED
#pragma weak pmpi_finalized = PMPI_FINALIZED
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_FINALIZED = pmpi_finalized__
#pragma weak pmpi_finalized_ = pmpi_finalized__
#pragma weak pmpi_finalized = pmpi_finalized__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_FINALIZED = pmpi_finalized_
#pragma weak pmpi_finalized__ = pmpi_finalized_
#pragma weak pmpi_finalized = pmpi_finalized_
#else
#pragma weak PMPI_FINALIZED = pmpi_finalized
#pragma weak pmpi_finalized__ = pmpi_finalized
#pragma weak pmpi_finalized_ = pmpi_finalized
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_finalized_ PMPI_FINALIZED
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_finalized_ pmpi_finalized__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_finalized_ pmpi_finalized
#else
#define mpi_finalized_ pmpi_finalized_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Finalized
#define MPI_Finalized PMPI_Finalized 

#else

#ifdef F77_NAME_UPPER
#define mpi_finalized_ MPI_FINALIZED
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_finalized_ mpi_finalized__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_finalized_ mpi_finalized
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_finalized_ ( MPI_Fint *v1, MPI_Fint *ierr ){
    int l1;
    *ierr = MPI_Finalized( &l1 );
    if (*ierr == MPI_SUCCESS) *v1 = MPIR_TO_FLOG(l1);
}
