/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2006 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */
#ifndef MPIU_OS_WRAPPERS_H_INCLUDED
#define MPIU_OS_WRAPPERS_H_INCLUDED
/* #include "mpiu_os_wrappers_conf.h" */
#include "mpiu_util_wrappers.h"
#include "mpiu_process_wrappers.h"
#include "mpiu_shm_wrappers.h"
#include "mpiu_sock_wrappers.h"
#endif /* MPIU_OS_WRAPPERS_H_INCLUDED */
