/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_ALLTOALLV( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_alltoallv__( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_alltoallv( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_alltoallv_( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_ALLTOALLV = PMPI_ALLTOALLV
#pragma weak mpi_alltoallv__ = PMPI_ALLTOALLV
#pragma weak mpi_alltoallv_ = PMPI_ALLTOALLV
#pragma weak mpi_alltoallv = PMPI_ALLTOALLV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_ALLTOALLV = pmpi_alltoallv__
#pragma weak mpi_alltoallv__ = pmpi_alltoallv__
#pragma weak mpi_alltoallv_ = pmpi_alltoallv__
#pragma weak mpi_alltoallv = pmpi_alltoallv__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_ALLTOALLV = pmpi_alltoallv_
#pragma weak mpi_alltoallv__ = pmpi_alltoallv_
#pragma weak mpi_alltoallv_ = pmpi_alltoallv_
#pragma weak mpi_alltoallv = pmpi_alltoallv_
#else
#pragma weak MPI_ALLTOALLV = pmpi_alltoallv
#pragma weak mpi_alltoallv__ = pmpi_alltoallv
#pragma weak mpi_alltoallv_ = pmpi_alltoallv
#pragma weak mpi_alltoallv = pmpi_alltoallv
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_ALLTOALLV( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPI_ALLTOALLV = PMPI_ALLTOALLV
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_alltoallv__( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_alltoallv__ = pmpi_alltoallv__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_alltoallv( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_alltoallv = pmpi_alltoallv
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_alltoallv_( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_alltoallv_ = pmpi_alltoallv_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_ALLTOALLV  MPI_ALLTOALLV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_alltoallv__  mpi_alltoallv__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_alltoallv  mpi_alltoallv
#else
#pragma _HP_SECONDARY_DEF pmpi_alltoallv_  mpi_alltoallv_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_ALLTOALLV as PMPI_ALLTOALLV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_alltoallv__ as pmpi_alltoallv__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_alltoallv as pmpi_alltoallv
#else
#pragma _CRI duplicate mpi_alltoallv_ as pmpi_alltoallv_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_ALLTOALLV( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_alltoallv__( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_alltoallv( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_alltoallv_( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_alltoallv__ = MPI_ALLTOALLV
#pragma weak mpi_alltoallv_ = MPI_ALLTOALLV
#pragma weak mpi_alltoallv = MPI_ALLTOALLV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_ALLTOALLV = mpi_alltoallv__
#pragma weak mpi_alltoallv_ = mpi_alltoallv__
#pragma weak mpi_alltoallv = mpi_alltoallv__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_ALLTOALLV = mpi_alltoallv_
#pragma weak mpi_alltoallv__ = mpi_alltoallv_
#pragma weak mpi_alltoallv = mpi_alltoallv_
#else
#pragma weak MPI_ALLTOALLV = mpi_alltoallv
#pragma weak mpi_alltoallv__ = mpi_alltoallv
#pragma weak mpi_alltoallv_ = mpi_alltoallv
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_ALLTOALLV( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_alltoallv__( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_alltoallv_( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_alltoallv( void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, void*, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_alltoallv__ = PMPI_ALLTOALLV
#pragma weak pmpi_alltoallv_ = PMPI_ALLTOALLV
#pragma weak pmpi_alltoallv = PMPI_ALLTOALLV
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_ALLTOALLV = pmpi_alltoallv__
#pragma weak pmpi_alltoallv_ = pmpi_alltoallv__
#pragma weak pmpi_alltoallv = pmpi_alltoallv__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_ALLTOALLV = pmpi_alltoallv_
#pragma weak pmpi_alltoallv__ = pmpi_alltoallv_
#pragma weak pmpi_alltoallv = pmpi_alltoallv_
#else
#pragma weak PMPI_ALLTOALLV = pmpi_alltoallv
#pragma weak pmpi_alltoallv__ = pmpi_alltoallv
#pragma weak pmpi_alltoallv_ = pmpi_alltoallv
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_alltoallv_ PMPI_ALLTOALLV
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_alltoallv_ pmpi_alltoallv__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_alltoallv_ pmpi_alltoallv
#else
#define mpi_alltoallv_ pmpi_alltoallv_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Alltoallv
#define MPI_Alltoallv PMPI_Alltoallv 

#else

#ifdef F77_NAME_UPPER
#define mpi_alltoallv_ MPI_ALLTOALLV
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_alltoallv_ mpi_alltoallv__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_alltoallv_ mpi_alltoallv
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_alltoallv_ ( void*v1, MPI_Fint *v2, MPI_Fint *v3, MPI_Fint *v4, void*v5, MPI_Fint *v6, MPI_Fint *v7, MPI_Fint *v8, MPI_Fint *v9, MPI_Fint *ierr ){

#ifndef HAVE_MPI_F_INIT_WORKS_WITH_C
    if (MPIR_F_NeedInit){ mpirinitf_(); MPIR_F_NeedInit = 0; }
#endif
    if (v1 == MPIR_F_MPI_IN_PLACE) v1 = MPI_IN_PLACE;
    *ierr = MPI_Alltoallv( v1, v2, v3, (MPI_Datatype)(*v4), v5, v6, v7, (MPI_Datatype)(*v8), (MPI_Comm)(*v9) );
}
