/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2008 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "hydra.h"
#include "hydt_ftb.h"

HYD_status HYDT_ftb_init(void)
{
    return HYD_SUCCESS;
}

HYD_status HYDT_ftb_finalize(void)
{
    return HYD_SUCCESS;
}

HYD_status HYDT_ftb_publish(const char *event_name, const char *event_payload)
{
    return HYD_SUCCESS;
}
