/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_FREE_MEM( void*, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_free_mem__( void*, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_free_mem( void*, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_free_mem_( void*, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_FREE_MEM = PMPI_FREE_MEM
#pragma weak mpi_free_mem__ = PMPI_FREE_MEM
#pragma weak mpi_free_mem_ = PMPI_FREE_MEM
#pragma weak mpi_free_mem = PMPI_FREE_MEM
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_FREE_MEM = pmpi_free_mem__
#pragma weak mpi_free_mem__ = pmpi_free_mem__
#pragma weak mpi_free_mem_ = pmpi_free_mem__
#pragma weak mpi_free_mem = pmpi_free_mem__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_FREE_MEM = pmpi_free_mem_
#pragma weak mpi_free_mem__ = pmpi_free_mem_
#pragma weak mpi_free_mem_ = pmpi_free_mem_
#pragma weak mpi_free_mem = pmpi_free_mem_
#else
#pragma weak MPI_FREE_MEM = pmpi_free_mem
#pragma weak mpi_free_mem__ = pmpi_free_mem
#pragma weak mpi_free_mem_ = pmpi_free_mem
#pragma weak mpi_free_mem = pmpi_free_mem
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_FREE_MEM( void*, MPI_Fint * );

#pragma weak MPI_FREE_MEM = PMPI_FREE_MEM
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_free_mem__( void*, MPI_Fint * );

#pragma weak mpi_free_mem__ = pmpi_free_mem__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_free_mem( void*, MPI_Fint * );

#pragma weak mpi_free_mem = pmpi_free_mem
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_free_mem_( void*, MPI_Fint * );

#pragma weak mpi_free_mem_ = pmpi_free_mem_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_FREE_MEM  MPI_FREE_MEM
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_free_mem__  mpi_free_mem__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_free_mem  mpi_free_mem
#else
#pragma _HP_SECONDARY_DEF pmpi_free_mem_  mpi_free_mem_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_FREE_MEM as PMPI_FREE_MEM
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_free_mem__ as pmpi_free_mem__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_free_mem as pmpi_free_mem
#else
#pragma _CRI duplicate mpi_free_mem_ as pmpi_free_mem_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_FREE_MEM( void*, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_free_mem__( void*, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_free_mem( void*, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_free_mem_( void*, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_free_mem__ = MPI_FREE_MEM
#pragma weak mpi_free_mem_ = MPI_FREE_MEM
#pragma weak mpi_free_mem = MPI_FREE_MEM
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_FREE_MEM = mpi_free_mem__
#pragma weak mpi_free_mem_ = mpi_free_mem__
#pragma weak mpi_free_mem = mpi_free_mem__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_FREE_MEM = mpi_free_mem_
#pragma weak mpi_free_mem__ = mpi_free_mem_
#pragma weak mpi_free_mem = mpi_free_mem_
#else
#pragma weak MPI_FREE_MEM = mpi_free_mem
#pragma weak mpi_free_mem__ = mpi_free_mem
#pragma weak mpi_free_mem_ = mpi_free_mem
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_FREE_MEM( void*, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_free_mem__( void*, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_free_mem_( void*, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_free_mem( void*, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_free_mem__ = PMPI_FREE_MEM
#pragma weak pmpi_free_mem_ = PMPI_FREE_MEM
#pragma weak pmpi_free_mem = PMPI_FREE_MEM
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_FREE_MEM = pmpi_free_mem__
#pragma weak pmpi_free_mem_ = pmpi_free_mem__
#pragma weak pmpi_free_mem = pmpi_free_mem__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_FREE_MEM = pmpi_free_mem_
#pragma weak pmpi_free_mem__ = pmpi_free_mem_
#pragma weak pmpi_free_mem = pmpi_free_mem_
#else
#pragma weak PMPI_FREE_MEM = pmpi_free_mem
#pragma weak pmpi_free_mem__ = pmpi_free_mem
#pragma weak pmpi_free_mem_ = pmpi_free_mem
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_free_mem_ PMPI_FREE_MEM
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_free_mem_ pmpi_free_mem__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_free_mem_ pmpi_free_mem
#else
#define mpi_free_mem_ pmpi_free_mem_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Free_mem
#define MPI_Free_mem PMPI_Free_mem 

#else

#ifdef F77_NAME_UPPER
#define mpi_free_mem_ MPI_FREE_MEM
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_free_mem_ mpi_free_mem__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_free_mem_ mpi_free_mem
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_free_mem_ ( void*v1, MPI_Fint *ierr ){
    *ierr = MPI_Free_mem( v1 );
}
