/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_OP_COMMUTATIVE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_op_commutative__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_op_commutative( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_op_commutative_( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_OP_COMMUTATIVE = PMPI_OP_COMMUTATIVE
#pragma weak mpi_op_commutative__ = PMPI_OP_COMMUTATIVE
#pragma weak mpi_op_commutative_ = PMPI_OP_COMMUTATIVE
#pragma weak mpi_op_commutative = PMPI_OP_COMMUTATIVE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_OP_COMMUTATIVE = pmpi_op_commutative__
#pragma weak mpi_op_commutative__ = pmpi_op_commutative__
#pragma weak mpi_op_commutative_ = pmpi_op_commutative__
#pragma weak mpi_op_commutative = pmpi_op_commutative__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_OP_COMMUTATIVE = pmpi_op_commutative_
#pragma weak mpi_op_commutative__ = pmpi_op_commutative_
#pragma weak mpi_op_commutative_ = pmpi_op_commutative_
#pragma weak mpi_op_commutative = pmpi_op_commutative_
#else
#pragma weak MPI_OP_COMMUTATIVE = pmpi_op_commutative
#pragma weak mpi_op_commutative__ = pmpi_op_commutative
#pragma weak mpi_op_commutative_ = pmpi_op_commutative
#pragma weak mpi_op_commutative = pmpi_op_commutative
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_OP_COMMUTATIVE( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPI_OP_COMMUTATIVE = PMPI_OP_COMMUTATIVE
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_op_commutative__( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_op_commutative__ = pmpi_op_commutative__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_op_commutative( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_op_commutative = pmpi_op_commutative
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_op_commutative_( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_op_commutative_ = pmpi_op_commutative_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_OP_COMMUTATIVE  MPI_OP_COMMUTATIVE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_op_commutative__  mpi_op_commutative__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_op_commutative  mpi_op_commutative
#else
#pragma _HP_SECONDARY_DEF pmpi_op_commutative_  mpi_op_commutative_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_OP_COMMUTATIVE as PMPI_OP_COMMUTATIVE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_op_commutative__ as pmpi_op_commutative__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_op_commutative as pmpi_op_commutative
#else
#pragma _CRI duplicate mpi_op_commutative_ as pmpi_op_commutative_
#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK) && \
    defined(USE_ONLY_MPI_NAMES)
extern FORT_DLL_SPEC void FORT_CALL MPI_OP_COMMUTATIVE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_op_commutative__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_op_commutative( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_op_commutative_( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_op_commutative__ = MPI_OP_COMMUTATIVE
#pragma weak mpi_op_commutative_ = MPI_OP_COMMUTATIVE
#pragma weak mpi_op_commutative = MPI_OP_COMMUTATIVE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_OP_COMMUTATIVE = mpi_op_commutative__
#pragma weak mpi_op_commutative_ = mpi_op_commutative__
#pragma weak mpi_op_commutative = mpi_op_commutative__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_OP_COMMUTATIVE = mpi_op_commutative_
#pragma weak mpi_op_commutative__ = mpi_op_commutative_
#pragma weak mpi_op_commutative = mpi_op_commutative_
#else
#pragma weak MPI_OP_COMMUTATIVE = mpi_op_commutative
#pragma weak mpi_op_commutative__ = mpi_op_commutative
#pragma weak mpi_op_commutative_ = mpi_op_commutative
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS) && defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_OP_COMMUTATIVE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_op_commutative__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_op_commutative_( MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_op_commutative( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_op_commutative__ = PMPI_OP_COMMUTATIVE
#pragma weak pmpi_op_commutative_ = PMPI_OP_COMMUTATIVE
#pragma weak pmpi_op_commutative = PMPI_OP_COMMUTATIVE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_OP_COMMUTATIVE = pmpi_op_commutative__
#pragma weak pmpi_op_commutative_ = pmpi_op_commutative__
#pragma weak pmpi_op_commutative = pmpi_op_commutative__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_OP_COMMUTATIVE = pmpi_op_commutative_
#pragma weak pmpi_op_commutative__ = pmpi_op_commutative_
#pragma weak pmpi_op_commutative = pmpi_op_commutative_
#else
#pragma weak PMPI_OP_COMMUTATIVE = pmpi_op_commutative
#pragma weak pmpi_op_commutative__ = pmpi_op_commutative
#pragma weak pmpi_op_commutative_ = pmpi_op_commutative
#endif /* Test on name mapping */
#endif /* Use multiple pragma weak */

#ifdef F77_NAME_UPPER
#define mpi_op_commutative_ PMPI_OP_COMMUTATIVE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_op_commutative_ pmpi_op_commutative__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_op_commutative_ pmpi_op_commutative
#else
#define mpi_op_commutative_ pmpi_op_commutative_
#endif /* Test on name mapping */

/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_Op_commutative
#define MPI_Op_commutative PMPI_Op_commutative 

#else

#ifdef F77_NAME_UPPER
#define mpi_op_commutative_ MPI_OP_COMMUTATIVE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_op_commutative_ mpi_op_commutative__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_op_commutative_ mpi_op_commutative
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_op_commutative_ ( MPI_Fint *v1, MPI_Fint *v2, MPI_Fint *ierr ){
    int l2;
    *ierr = MPI_Op_commutative( (MPI_Op)*v1, &l2 );
    if (*ierr == MPI_SUCCESS) *v2 = MPIR_TO_FLOG(l2);
}
